const bs58 = require('bs58');
import {
  MangoInstructionLayout,
  IDS,
  PerpMarket,
  Config,
  Cluster,
  GroupConfig,
  awaitTransactionSignatureConfirmation,
  PerpMarketConfig,
} from '@blockworks-foundation/mango-client';
import { jsonParser } from './jsonParsers';
import { anchorParser } from './anchorParsers';

export function parseTransactions(transactionsResult, mangoProgramId) {
  let processStates: any = [];

  let parsedTransactions: any = {
    deposits_withdraws: [],

    cache_prices: [],
    cache_indexes: [],

    settle_pnl: [],
    settle_fees: [],
    // 'force_settle_quote_positions': [], // Deprecated?

    liquidate_token_and_token: [],
    liquidate_token_and_perp: [],
    liquidate_perp_market: [],

    token_bankruptcy: [],
    perp_bankruptcy: [],

    fill_events: [],

    net_balances: [],
    redeem_mngo: [],
    funding: [],

    open_orders_balances: [],
    mango_accrual: [],
    // Mango: PlacePerpOrder
    // Mango: CancelPerpOrderByClientI
    // Mango: CancelPerpOrder
  };

  for (let transactionResult of transactionsResult) {
    let [transactionJson, signature] = transactionResult;
    try {
      let transaction = JSON.parse(transactionJson);
      let result = transaction.result;

      if (result.meta.err !== null) {
        processStates.push({
          signature: signature,
          process_state: 'transaction error',
        });
      } else {
        let slot = result.slot;
        let blockTime = result.blockTime;
        let blockDatetime = new Date(blockTime * 1000).toISOString();

        // only look at cases where instruction is for mango program id
        // Skip attempting to parse serum instructions, etc
        // @clarkeni - what if someone calls mango program via CPI? Do the instruction list contain inner instructions
        let instructions = result.transaction.message.instructions.filter(
          (ix) =>
            ix.programId === mangoProgramId &&
            !(
              // Old mango group - not in ids.json so have to hardcode here to ignore
              (
                ix.accounts &&
                ix.accounts[0] ===
                  '2WNLfEMzhgwBPn6QptT43SdZy9cXTUbVJCMdCfimg4oi'
              )
            ),
        );
        
        // Anchor logging was deployed at slot 100936906 - use different parsing code before and after this slot 
        const ancorDeploymentSlot = 100936906
        if (slot < ancorDeploymentSlot) {

          // Populate instruction num and name for each instruction
          let ixNum = 1;
          for (const ix of instructions) {
            let decodeData = bs58.decode(ix.data);
            let decodedInstruction = MangoInstructionLayout.decode(decodeData, 0);
            let instructionName = Object.keys(decodedInstruction)[0];
            ix.instructionNum = ixNum;
            ix.instructionName = instructionName;
            ixNum++;
          }
          
          jsonParser(parsedTransactions, result, instructions, signature, blockTime, slot, blockDatetime)
        } else {
          anchorParser(parsedTransactions, result, signature, blockTime, slot, blockDatetime)
        }
        
        processStates.push({
          signature: signature,
          process_state: 'processed',
        });
      }
    } catch (e: any) {
      console.log(e.stack);
      processStates.push({
        signature: signature,
        process_state: 'parsing error',
      });
    }
  }

  return [processStates, parsedTransactions];
}
